#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/wait.h>
#include <linux/unistd.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <regex.h>
#include <fcntl.h>
#include <time.h>
#include <sys/mman.h>
#include <dlfcn.h>
#include <popt.h>
#include <sys/socket.h>
#include <ctype.h>
#include <glib.h>

#include "tuxmodule.h"

#if defined(__ia64__)
#define __NR_tux 1120
inline int tux(unsigned int action, user_req_t *req)
{
   return syscall(__NR_tux,action,req);
}

#else

#if defined(__powerpc__)
#define __NR_tux 225
#elif defined(__x86_64__)
#define __NR_tux 184
#elif defined(__alpha__)
#define __NR_tux 397
#elif (defined(__s390__) || defined (__s390x__))
#define __NR_tux 242
#elif (defined (__i386__) || defined (__arm__))
#define __NR_tux 222
#else
#warning unsupported architecture, guessing __NR_tux=222 like x86...
#define __NR_tux 222
#endif

_syscall2 (int, tux, unsigned int, action, user_req_t *, req)
#endif


char *TUXAPI_version;
unsigned long TUXAPI_version_len;
char *TUXAPI_docroot;
unsigned long TUXAPI_docroot_len;
char *TUXAPI_servername;
unsigned long TUXAPI_servername_len;

#define TUX_DEBUG 0

#if TUX_DEBUG
# define Dprintk(x...) do { printf(x); fflush(stdout); } while (0)
#else
# define Dprintk(x...) do { } while (0)
#endif

#define TUX_TRACE(req,info) Dprintk("TUX trace at %s:%d, request %p, info %d, event %d (meth:%d, query:%s, cookies:%s).\n", __FILE__, __LINE__, (req)->id, info, (req)->event, (req)->http_method, (req)->query, (req)->cookies_len ? (req)->cookies : "<no cookies>")

#define MAX_NR_THREADS 8

static int nr_threads;
static void *shared_buf;
static int shared_idx = 0;

static int max_module_idx = 1;
static int (**handle_events_array) (user_req_t *req);
char **TUXAPI_modulename_array;

void die(int exitcode, char *format, ...) {
	va_list ap;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	exit(exitcode);
}

void do_regerror(int errcode, const regex_t *preg) {
	char *errbuf;
	size_t errbuf_size;

	errbuf_size = regerror(errcode, preg, NULL, 0); /* missing ; */
	errbuf = alloca(errbuf_size);
	if(!errbuf) {
		perror("alloca");
		return;
	}

	regerror(errcode, preg, errbuf, errbuf_size);
	fprintf(stderr, "%s\n", errbuf);
}

void * TUXAPI_malloc_shared (unsigned int len)
{
	void *ret;

	ret = (void *)((char *)shared_buf + shared_idx);
	shared_idx += len;

	return ret;
}

char * TUXAPI_alloc_read_objectbuf (user_req_t *req, int fd)
{
	char *buf;

	buf = malloc(req->objectlen);
	read(fd, buf, req->objectlen);
	return buf;
}

void TUXAPI_free_objectbuf (user_req_t *req, void *buf)
{
	free(buf);
}

void handle_all_events (int thread_nr)
{
	int ret = 0;
	user_req_t req_s;
	user_req_t *req = &req_s;
	int (*handle_events) (user_req_t *req);

	memset(req, 0, sizeof(*req));
	req->thread_nr = thread_nr;
	req->module_index = 0;

event_loop:
/**/	TUX_TRACE(req, req->event);
/**/	TUX_TRACE(req, ret);
	ret = tux(TUX_ACTION_EVENTLOOP, req);

	if (ret == TUX_RETURN_USERSPACE_REQUEST) {
		do {
			handle_events = handle_events_array[req->module_index];
			if (!handle_events)
				BUG();
			ret = handle_events(req);
		} while (ret == TUX_RETURN_USERSPACE_REQUEST);
	}
	if (ret == TUX_RETURN_SIGNAL)
		goto event_loop;
	printf("last event: %d.\n", ret);
}

void initialize_mimetypes() {
	user_req_t req_s;
	user_req_t *req = &req_s;
	int mimetypes;
	struct stat sb;
	char *contents;
	char *confline, *confnext;
	char *cache_control;

	/* these will have to become static if we make mimetypes dynamic */
	GHashTable *exts;

	regex_t p;
	regmatch_t *pmatch;
	int rcomp_err, rexec_err;

	exts = g_hash_table_new(g_str_hash, g_str_equal);

	memset(req, 0, sizeof(*req));

	mimetypes = open("/etc/tux.mime.types", O_RDONLY);
	if (mimetypes == -1)
		die(1, "Missing /etc/tux.mime.types file, exiting\n");
	fstat(mimetypes, &sb);
	contents = alloca(sb.st_size);
	read(mimetypes, contents, sb.st_size);
	close(mimetypes);

	rcomp_err = regcomp (&p, "([^ \t][^ \t]*)[ \t][ \t]*(.*)",
			     REG_EXTENDED|REG_NEWLINE);
	if (rcomp_err) do_regerror(rcomp_err, &p);

	pmatch = alloca(sizeof(regmatch_t) * (p.re_nsub+1));

	for (confline = contents; confline; confline = confnext) {
		char *extensions, *extension, *nextext;
		char *mimetype;
		int mimelen, extlen;

		confnext = strchr(confline, '\n');
		if (confnext) *confnext++ = '\0';

		if (confline[0] == '#') continue;

		rexec_err = regexec(&p, confline, p.re_nsub+1, pmatch, 0);
		if (rexec_err) continue;

		/* extension(s) listed as well as a mime type */
		mimelen = pmatch[1].rm_eo - pmatch[1].rm_so;
		mimetype = malloc(mimelen+1);
		extlen = pmatch[2].rm_eo - pmatch[2].rm_so;
		extensions = malloc(extlen+1);

		strncpy(mimetype, confline+pmatch[1].rm_so, mimelen);
		mimetype[mimelen] = '\0';
		strncpy(extensions, confline+pmatch[2].rm_so, extlen);
		extensions[extlen] = '\0';

#if (INTERFACE_VERSION >= 3)
		req->object_addr = (uint64_t)(long)mimetype;
#else
		req->object_addr = mimetype;
#endif

		extension = extensions;
		cache_control = NULL;
		for (nextext = extension; *nextext; extension = nextext) {
			while ((*nextext != ' ') &&
			       (*nextext != '\t') &&
			       (*nextext != '\0')) {
				if (*nextext == '|') {
					cache_control = nextext+1;
					*nextext = '\0';
				}
				nextext++;
			}
			while ((*nextext == ' ') ||
			       (*nextext == '\t'))
				*nextext++ = '\0';

			if (g_hash_table_lookup(exts, extension)) goto nextline;
			g_hash_table_insert(exts, strdup(extension), (void *)1);

			req->objectname[0] = '.';

			strncpy(req->objectname+1, extension, MAX_URI_LEN);
			if (cache_control)
				strncpy(req->cache_control, cache_control, MAX_URI_LEN);
			else
				req->cache_control[0]='\0';
			tux(TUX_ACTION_REGISTER_MIMETYPE, req);
		}
		
nextline:
		free(mimetype);
		free(extensions);
	}
	regfree(&p);
	g_hash_table_destroy(exts);
}

void tux_time (int ignored) {
	return;
}

static void replace_commandline (int argc, char **argv, const char *new)
{
	int len = strlen(argv[argc-1]);

	// tell the kernel that we have a new commandline
	argv[argc-1][len] = 1;

	sprintf(argv[0], new);
}

static pid_t tux_start_date_timer (int interval, int argc, char **argv) {
	pid_t child;
	struct itimerval it;
	struct sigaction sa;

	if ((child = fork()))
		return child;

	replace_commandline(argc, argv, "[TUX date]");
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = tux_time;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;

	sigaction(SIGALRM, &sa, NULL);

	memset(&it, 0, sizeof(it));
	it.it_interval.tv_sec = interval;
	it.it_value.tv_sec = interval;
	setitimer(ITIMER_REAL, &it, NULL);

	while (1) {
const char weekday[][7] = { "Sun", "Mon",  "Tue", "Wed", "Thu", "Fri", "Sat" };
const char month[][12] = { "Jan", "Feb", "Mar", "Apr", "May",
			"Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

		user_req_t req;
		time_t t;
		struct tm *tm;

		pause();
		time(&t);

		memset(&req, 0, sizeof(req));
		tm = gmtime(&t);
		/*
		 * Faster date construction:
		 *
		 * RFC 1123 time "Sun, 06 Nov 1994 08:49:37 GMT"
		 */
		sprintf(req.new_date,
			"%s, %02d %s %04d %02d:%02d:%02d GMT",
				weekday[tm->tm_wday], tm->tm_mday,
				month[tm->tm_mon], 1900+tm->tm_year,
				tm->tm_hour, tm->tm_min, tm->tm_sec);
		tux(TUX_ACTION_CURRENT_DATE, &req);
	}
	return 0;
}

uid_t get_uid(char *name) {
	char *t = name;

	while (isdigit(*t++)); /* pass over any digits */

	if (*t) {
		/* name is not comprised wholely of digits, must be a name */
		struct passwd *p;
		p = getpwnam(name);
		if (!p) die(1, "Could not find user name \"%s\"\n", name);
		return p->pw_uid;
	} else {
		/* name must be all digits */
		return atoi(name);
	}
}

gid_t get_gid(char *name) {
	char *t = name;

	while (isdigit(*t++)); /* pass over any digits */

	if (*t) {
		/* name is not comprised wholely of digits, must be a name */
		struct group *g;
		g = getgrnam(name);
		if (!g) die(1, "Could not find group name \"%s\"\n", name);
		return g->gr_gid;
	} else {
		/* name must be all digits */
		return atoi(name);
	}
}

void tux_stop_date_timer (pid_t child) {
	kill (child, SIGTERM);
	waitpid(child, NULL, 0);
}

void tux_stop(void) {
	user_req_t req; 
	int ret;

	memset(&req, 0, sizeof(req));
	ret = tux(TUX_ACTION_SHUTDOWN, &req);

	exit(0);
}

#define TUX_VERSION "TUX/2.0 (Linux, Red Hat)"

int main (int argc, char **argv)
{
	int pid [MAX_NR_THREADS];
	pid_t timer = 0;
	void *handle;
	int ret, i;
	int version, major, minor, patchlevel;
	int mod_idx = 1;

	char *module_path = "";
	int do_daemon = 0;
	int stop_tux = 0;
	int date_interval=1;
	char *group_id_string=NULL, *user_id_string=NULL;
	gid_t group_id;
	uid_t user_id;
	signed char c;
	poptContext optCon;

	struct poptOption optionsTable[] = {
	    { "threads", 't', POPT_ARG_INT, &nr_threads, 0,
	      "number of tux threads", "threads" },
	    { "docroot", 'r', POPT_ARG_STRING, &TUXAPI_docroot, 0,
	      "directory path for document root", "path" },
	    { "modpath", 'm', POPT_ARG_STRING, &module_path, 0,
	      "directory path for TUXAPI user-space modules", "path" },
	    { "userid", 'u', POPT_ARG_STRING, &user_id_string, 0,
	      "user ID to use for modules (nobody)", "uid|name" },
	    { "groupid", 'g', POPT_ARG_STRING, &group_id_string, 0,
	      "group ID to use for modules (nobody)", "gid|name" },
	    { "daemon", 'd', POPT_ARG_NONE, &do_daemon, 0,
	      "run in the background as a daemon", NULL },
	    { "date-interval", 'D', POPT_ARG_INT, &date_interval, 0,
	      "how often to update the date string (1)", "seconds" },
	    { "stop", 's', POPT_ARG_NONE, &stop_tux, 0,
	      "stop the currently-runnning TUX system", NULL },
	      POPT_AUTOHELP
	    { NULL, 0, 0, NULL, 0 }
	};

	optCon = poptGetContext("tux", argc, (const char **)argv,
				optionsTable, 0);
	while ((c = poptGetNextOpt(optCon)) >= 0);

	if (c < -1) /* an error occurred during option processing */
		die(1, "%s: %s\n", 
			poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
			poptStrerror(c));

	if (stop_tux) tux_stop();

	user_id = get_uid(user_id_string ? user_id_string : "nobody");
	group_id = get_gid(group_id_string ? group_id_string : "nobody");

	TUXAPI_version = (char *)malloc(sizeof(TUX_VERSION));
	memcpy(TUXAPI_version, TUX_VERSION, sizeof(TUX_VERSION));

	version = tux(TUX_ACTION_QUERY_VERSION, NULL);
	if (errno == ENOSYS) {
	    die(1, "TUX not implemented in this kernel\n");
	}
	if (version < 0) {
	    /* not implemented ==> major 2 */
	    major = 2;
	    minor = 0;
	    patchlevel = 0;
	} else {
	    /* major supported by kernel is in top 16 bits of return code */
	    major = version >> 24;
	    minor = (version >> 16) & 0xff;
	    patchlevel = version & 0xffff;
	}
	fprintf(stderr, "Tux version: %d/%d/%d\n", major, minor, patchlevel);
	if (major != TUX_MAJOR_VERSION)
	    die(1, "TUX protocol version mismatch, kernel %d != daemon %d\n",
		   major, TUX_MAJOR_VERSION);

	TUXAPI_servername = (char *)malloc(100);
	if (gethostname(TUXAPI_servername, 100) < 0)
		die(1, "TUX hostname problem!\n");
	TUXAPI_servername_len = strlen(TUXAPI_servername);

	initialize_mimetypes();

	/* timer must run as root, start with root privs */
	timer = tux_start_date_timer(date_interval, argc, argv);

	#define MMAP_FILE "/var/tux/tmp_mmap"
	{
		int fd;
		char zerobuff [4096] = { 0, };

		fd = open(MMAP_FILE, O_RDWR|O_CREAT|O_TRUNC, 0700);
		if (fd < 0)
			die(1, "could not open mmap file, does the /var/tux/ directory exist?\n");
		write (fd, zerobuff, 4096);
		shared_buf = (double *)mmap(0, 4096, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
		close(fd);
		memset(shared_buf, 0, 4096);
	}

	/* We need to better handle the no-modules case; we'll segfault
	   right now if something comes through as a USER_REQUEST.
	   FIXME!
	 */
	if (poptPeekArg(optCon)) {
		const char *modulename;

		handle_events_array = malloc(sizeof(*handle_events_array) * (argc-3));
		TUXAPI_modulename_array = malloc(sizeof(*TUXAPI_modulename_array) * (argc-3));
		if (!handle_events_array || !TUXAPI_modulename_array) {
			tux_stop_date_timer(timer);
			die(1, "%s: no RAM", strerror(errno));
		}

		while ((modulename = poptGetArg(optCon))) {
			user_req_t req;
			int (*mod_handle_events) (user_req_t *req);
			int *mod_protocol_version_p;
			int mod_protocol_version;
			void (*mod_init) (void);
			char *libname;
			unsigned int modulename_len = strlen(modulename);

			if (modulename_len >= MAX_URI_LEN)
				die(1, "module name longer than %d characters!", MAX_URI_LEN);
			libname = malloc(strlen(module_path)+modulename_len+11);
			strcpy(libname, module_path);
			strcat(libname, "/");
			strcat(libname, modulename);
			strcat(libname, ".so");
			printf("opening shared library %s\n", libname);
			/* FIXME: check ownership and permissions first? */
			handle = dlopen (libname, RTLD_LAZY);
			printf ("handle: %p\n", handle);
			if (!handle) {
				tux_stop_date_timer(timer);
				die(1, "%s: dlopen() handle missing", dlerror());
			}

			/* make sure protocol version is right */
			mod_protocol_version_p = dlsym(handle, "TUXAPI_protocol_version");
			if (mod_protocol_version_p) {
			    mod_protocol_version = *mod_protocol_version_p;
			} else {
			    mod_protocol_version = 2;
			}
			if (mod_protocol_version != TUX_MAJOR_VERSION) {
				tux_stop_date_timer(timer);
				die(1, "module version mismatch:"
				       "module version %d does not match kernel version %d",
				       mod_protocol_version, TUX_MAJOR_VERSION);
			}

			mod_handle_events = dlsym(handle, "TUXAPI_handle_events");
			if (!mod_handle_events) {
				tux_stop_date_timer(timer);
				die(1, "%s: TUXAPI_handle_events() function missing", strerror(errno));
			}
			mod_init = dlsym(handle, "TUXAPI_init");

			printf ("TUXAPI_init: %p\n", mod_init);
			printf ("TUXAPI_handle_events: %p\n", mod_handle_events);

			memset(&req, 0, sizeof(req));
			req.module_index = mod_idx;
			strcpy(req.objectname, modulename);
			req.version_major = TUX_MAJOR_VERSION;
			req.version_minor = TUX_MINOR_VERSION;
			req.version_patch = TUX_PATCHLEVEL_VERSION;

			ret = tux(TUX_ACTION_REGISTER_MODULE, &req);
			printf("register {%s} - TUX returned %d...\n", req.objectname, ret);

			handle_events_array[mod_idx] = mod_handle_events;
			TUXAPI_modulename_array[mod_idx] = malloc(strlen(modulename)+1);
			if (!TUXAPI_modulename_array[mod_idx]) {
				tux_stop_date_timer(timer);
				die(1, "%s: malloc problem", strerror(errno));
			}
			strcpy(TUXAPI_modulename_array[mod_idx], modulename);

			strcat(libname, modulename);
			max_module_idx++;
			mod_idx++;
			if (mod_init)
				mod_init();
		}
	}
	poptFreeContext(optCon);

	ret = tux(TUX_ACTION_STARTUP, NULL);
	if (ret) {
		tux_stop_date_timer(timer);
		die(1, "could not initialize TUX!\n");
	}
	TUXAPI_docroot_len = strlen(TUXAPI_docroot);

	if (do_daemon)
		daemon(1, 0);

	for (i = 0; i < nr_threads; i++) {
		user_req_t req;
		char commandline [100];

		pid[i] = fork();
		if (pid[i] < 0) {
			tux_stop_date_timer(timer);
			die (1, "could not fork TUX thread %d!\n", i);
		}
		if (pid[i])
			continue;

		sprintf(commandline, "[TUX worker %d]", i);
		replace_commandline(argc, argv, commandline);

		req.thread_nr = i;
		ret = tux(TUX_ACTION_STARTTHREAD, &req);
		printf("start - TUX %d returned %d...\n", i, ret);
		if (ret)
			die(1, "could not start TUX thread %d!\n", i);

		/* Drop permissions */
		if (group_id) setgid(group_id);
		if (user_id) setuid(user_id);

		chdir("/");

		handle_all_events(i);

		exit(0);
	}
	replace_commandline(argc, argv, "[TUX manager]");

	ret = 0;
	for (i = 0; i < nr_threads; i++) {
		int status = 0;

		printf("waiting for TUX thread %d.\n", i);
		waitpid(pid[i], &status, 0);
		printf("TUX thread %d returned with status %d.\n", i, status);
		if (status)
			ret = 1;
	}

	tux_stop_date_timer(timer);

	return ret;
}

